
/*
 * dis_c84.c -- disassembler for PIC 16C84, input is a INHX8M object file.
 *		Output to stdout.
 *
 * Copyright (C) 1995-6 James Cleverdon
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Address:
 *	James Cleverdon
 *	16465 S.W. Estuary Dr. # 101
 *	Beaverton, OR 97006-7954, USA
 *
 *	jamesc@sequent.com
 *
 * $Log:	dis_c84.c,v $
 * Revision 1.1  96/01/25  00:51:11  jamesc
 * Made disassembly format more like the assembler's
 * listing output.
 * 
 * Revision 1.0  1995/12/19 01:28:34  jamesc
 * "CSE510_project"
 *
 */

#ident "$Id: dis_c84.c,v 1.1 96/01/25 00:51:11 jamesc Exp Locker: jamesc $"

#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "c84_mem.h"

#define MAXLINE 	256		/* maximum line length	*/

#define NEL(x)		(sizeof(x) / sizeof(*x))

extern int		errno;

static short		code_mem[CODE_MAX + 1];
static short		eedata_mem[EEDATA_MAX + 1];
static short		id_mem[ID_MAX + 1];
static short		fuse_mem;

static char		line[MAXLINE];

#define CLASS_SHIFT	12		/* rshift by 12 to get class	*/
#define CL_MISC 	0		/* miscellaneous opcodes	*/
#define CL_BIT		1		/* bit field opcodes		*/
#define CL_ADDR 	2		/* goto and call opcodes	*/
#define CL_LIT		3		/* literal operator opcodes	*/

#define DEST_MASK	0x80		/* mask for destination code	*/
#define FILE_MASK	0x7F		/* mask for file register addr	*/

#define MISC_T_SHIFT	8		/* rshift by 8 and mask for type */
#define MISC_T_MASK	0xF		/* mask off file,dest op types	*/
#define F_D_D_SHIFT	7		/* shift to get dest code field */
#define F_D_D_MASK	0x1		/* mask for dest code field	*/
#define F_D_T_SHIFT	8		/* shift to get dest code field */
#define F_D_T_MASK	0xF		/* mask for dest code field	*/
#define TRIS_MASK	0x7		/* tris register mask		*/
#define TRIS_VAL	0x60		/* tris opcode value		*/

#define BIT_B_SHIFT	7		/* rshift by 7 and mask for bit */
#define BIT_B_MASK	0x7		/* mask off bit number value	*/
#define BIT_T_SHIFT	10		/* rshift by 10 and mask for bit */
#define BIT_T_MASK	0x3		/* mask off bit op type 	*/

#define ADDR_GOTO	0x800		/* is a goto if this bit is set */
#define ADDR_MASK	0x7FF		/* address value mask		*/

#define LIT_T_SHIFT	8		/* rshift by 8 and mask for type */
#define LIT_T_MASK	0xF		/* mask off literal op types	*/
#define LIT_V_MASK	0xFF		/* mask for literal value	*/


static char	*bit_ops[] = { "BCF", "BSF", "BTFSC", "BTFSS" };

static char	*lit_ops[] = {
 "MOVLW", "MOVLW", "MOVLW", "MOVLW", "RETLW", "RETLW", "RETLW", "RETLW",
 "IORLW", "ANDLW", "XORLW", "?????", "SUBLW", "SUBLW", "ADDLW", "ADDLW"
};

static char	*f_d_ops[] = {
 "Bad_0", "Bad_1", "SUBWF", "DECF",  "IORWF", "ANDWF", "XORWF", "ADDWF",
 "MOVF",  "COMF",  "INCF",  "DECFSZ","RRF",   "RLF",   "SWAPF", "INCFSZ"
};

static char	*d_names[] = { "W", "F" };


static void
dis(register uint word)
{
	switch (word >> CLASS_SHIFT) {
	case CL_MISC:
		switch ((word >> MISC_T_SHIFT) & MISC_T_MASK) {
		case 0x0:
			if (word & DEST_MASK) {
				printf("MOVWF\t0x%02X\n", word & FILE_MASK);
				break;
			}
			switch (word) {
			case 0x00:
				printf("NOP\n");
				break;
			case 0x20:
			case 0x40:
			case 0x60:
				printf("NOP\t{%02X}\n", word);
				break;
			case 0x08:
				printf("RETURN\n");
				break;
			case 0x09:
				printf("RETFIE\n");
				break;
			case 0x62:
				printf("OPTION\n");
				break;
			case 0x63:
				printf("SLEEP\n");
				break;
			case 0x64:
				printf("CLRWDT\n");
				break;
			case 0x65:
			case 0x66:
				printf("TRIS\t%X\n", word & TRIS_MASK);
				break;
			default:
				printf("[%04X]\n", word);
				break;
			}
			break;
		case 0x1:
			if (word & DEST_MASK) {
				printf("CLRF\t0x%02X\n", word & FILE_MASK);
			}
			else {
				printf("CLRW\n");
			}
			break;
		default:
			printf("%s\t0x%02X, %s\n",
				f_d_ops[(word >> F_D_T_SHIFT) & F_D_T_MASK],
				word & FILE_MASK,
				d_names[(word >> F_D_D_SHIFT) & F_D_D_MASK]);
			break;
		}
		return;
	case CL_BIT:
		printf("%s\t0x%02X, %d\n",
			bit_ops[(word >> BIT_T_SHIFT) & BIT_T_MASK],
			word & FILE_MASK,
			(word >> BIT_B_SHIFT) & BIT_B_MASK);
		return;
	case CL_ADDR:
		printf("%s\t0x%03X\n",
			((word & ADDR_GOTO) ? "GOTO" : "CALL"),
			word & ADDR_MASK);
		return;
	case CL_LIT:
		printf("%s\t0x%02X\n",
			lit_ops[(word >> LIT_T_SHIFT) & LIT_T_MASK],
			word & LIT_V_MASK);
		return;
	default:
		printf("Hosed word value = 0x%08X\n", word);
		exit(1);
	}
}


/*
 * got_any -- return non-zero if any words are defined, else zero
 */

static int
got_any(register short *shp, register int num)
{
	while (--num >= 0) {
		if (*shp++ >= 0) {
			return (1);
		}
	}

	return (0);
}


static void
dis_all(void)
{
	register short	*shp;
	register int	addr;
	register int	num;
	static char	*osc_types[] = {
		"LP_OSC",	"XT_OSC",	"HS_OSC",	"RC_OSC"
	};
	static char	on_str[] = "ON";
	static char	off_str[] = "OFF";


	shp = code_mem;
	addr = CODE_START;
	num = NEL(code_mem);
	for ( ; --num >= 0; shp++, addr++) {
		if (*shp < 0)
			continue;
		printf("%04X:  %04X\t", addr, *shp);
		dis(*shp);
	}

	shp = eedata_mem;
	addr = EEDATA_START;
	num = NEL(eedata_mem);
	if (got_any(shp, num)) {
		printf("\nEEPROM data\n");

		for ( ; --num >= 0; shp++, addr++) {
			if (*shp < 0)
				continue;
			printf("%04X:E %02X\n", addr, *shp);
		}
	}

	shp = id_mem;
	addr = ID_START;
	num = NEL(id_mem);
	if (got_any(shp, num)) {
		printf("\nID data\n");

		for ( ; --num >= 0; shp++, addr++) {
			if (*shp < 0)
				continue;
			printf("%04X:I %04X\n", addr, *shp);
		}
	}

	if (fuse_mem >= 0) {
		printf("\nFuse data = 0x%02x [%s",
			fuse_mem, osc_types[fuse_mem & 0x3]);
		printf(" | WDTE_%s | PWRTE_%s | CP_%s]\n",
			((fuse_mem & 0x4) ? on_str : off_str),
			((fuse_mem & 0x8) ? on_str : off_str),
			((fuse_mem & 0x10) ? off_str : on_str));
	}

	(void) fflush(stdout);
}


/*
 * get_hex_byte -- quick 'n dirty two digit hex byte routine
 */

static int
get_hex_byte(register char *cp)
{
	register int	c;
	register int	val;
	register int	n;

	val = 0;
	n = 2;
	do {
		c = *cp++;
		val <<= 4;
		if (isdigit(c)) {
			val |= c - '0';
		} else if (isxdigit(c)) {
			if (islower(c))
				val |= c - 'a' + 10;
			else
				val |= c - 'A' + 10;
		} else {
			return (-1);
		}
	} while (--n > 0);

	return (val);
}


/*
 * load -- load INHX8M records into all four memory spaces
 */

static void
load(FILE *ifp)
{
	register int	line_num;
	register int	n;
	register int	v;
	register uint	chksum;
	register uint	val;
	register uint	mask;
	register short	*sp;
	register char	*cp;
	uint		nbytes;
	uint		addr;

	/* initialize all memory spaces to "unused" */

	(void) memset(code_mem, -1, sizeof(code_mem));
	(void) memset(eedata_mem, -1, sizeof(eedata_mem));
	(void) memset(id_mem, -1, sizeof(id_mem));
	fuse_mem = -1;

	line_num = 0;
	while (fgets(line, sizeof(line), ifp) != NULL) {
		++line_num;
		cp = line;
		if (*cp++ != ':') {
			continue;		/* skip comment lines */
		}

		if ((v = get_hex_byte(cp)) < 0) {		/* nbytes */
			goto bad_rec;
		}
		cp += 2;
		nbytes = v;
		chksum = v;

		/* only do words for 16C84 */

		if (nbytes & 1) {
			fprintf(stderr,
				"Odd Nbytes field = 0x%02X (line # %d)\n",
				nbytes, line_num);
			exit(1);
		}

		if ((v = get_hex_byte(cp)) < 0) {		/* addr high */
			goto bad_rec;
		}
		cp += 2;
		addr = v << 8;
		chksum += v;
		if ((v = get_hex_byte(cp)) < 0) {		/* addr low */
			goto bad_rec;
		}
		cp += 2;
		addr |= v;
		chksum += v;

		/* check address (the 2*addr is an artifact of INHX8M) */

		addr /= 2;
		if (addr >= CODE_START && addr <= CODE_END) {
			if (addr + nbytes/2 > CODE_END + 1) {
				goto bad_addr;
			}
			sp = &code_mem[addr - CODE_START];
			mask = ~0x3FFF;
		} else if (addr >= EEDATA_START && addr <= EEDATA_END) {
			if (addr + nbytes/2 > EEDATA_END + 1) {
				goto bad_addr;
			}
			sp = &eedata_mem[addr - EEDATA_START];
			mask = ~0x00FF;
		} else if (addr >= ID_START && addr <= ID_END) {
			if (addr + nbytes/2 > ID_END + 1) {
				goto bad_addr;
			}
			sp = &id_mem[addr - ID_START];
			mask = ~0x3FFF;
		} else if (addr == FUSE_ADDR) {
			if (nbytes > 2) {
				goto bad_addr;
			}
			sp = &fuse_mem;
			mask = ~0x3FFF;
		} else {
bad_addr:
			fprintf(stderr,
				"Bad Address field = 0x%04X (line # %d)\n",
				2 * addr, line_num);
			exit(1);
		}

		if ((v = get_hex_byte(cp)) < 0) {		/* type */
			goto bad_rec;
		}
		cp += 2;
		chksum += v;

		/* check type field */

		switch (v) {
		case 0:
			break;				/* data record	*/
		case 1:
			return;				/* end-of-data rec */
		default:
			fprintf(stderr,
				"Bad record type field = 0x%02X (line # %d)\n",
				v, line_num);
			exit(1);
		}

		/*
		 * read data words
		 */

		n = nbytes;
		while (--n >= 0) {
			if ((v = get_hex_byte(cp)) < 0) {	/* low byte */
				goto bad_data;
			}
			cp += 2;
			val = v;
			chksum += v;
			--n;

			if ((v = get_hex_byte(cp)) < 0) {	/* high byte */
				goto bad_data;
			}
			cp += 2;
			val |= v << 8;
			chksum += v;
			if (val & mask) {
				goto bad_word;
			}
			*sp++ = val;
		}

		/* finish checksum */

		if ((v = get_hex_byte(cp)) < 0) {		/* byte */
			fprintf(stderr,
				"Munged Checksum! (column # %d, line # %d)\n",
				(cp - line), line_num);
			exit(1);
		}
		chksum += v;
		if ((chksum &= 0xFF) != 0) {
			fprintf(stderr,
				"Bad Checksum = 0x%02X on line # %d!\n",
				chksum, line_num);
		}
	}

	return;

bad_rec:
	fprintf(stderr, "dis:  Munged Intel Hex record (line # %d)\n",
		line_num);
	exit(1);

bad_data:
	fprintf(stderr, "Bad Data! (column # %d, line # %d)\n",
		(cp - line), line_num);
	exit(1);

bad_word:
	fprintf(stderr, "Bad Word! 0x%04X (column # %d, line # %d)\n",
		val, (cp - line - 4), line_num);
	exit(1);
}


void
main(int argc, char **argv)
{
	FILE	*fp = stdin;
	char	*hex_file;

	if (argc > 1) {
		hex_file = argv[1];
		fp = fopen(hex_file, "r");
		if (!fp) {
			perror(hex_file);
			exit(1);
		}
	}

	load(fp);
	dis_all();

	exit(0);
}
